#include <u.h>
#include <libc.h>
#include <bio.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include "../lib/asn1.h"
#include "../lib/snmp_asn1.h"
#include "../lib/snmp.h"
#include "../lib/snmp_debug.h"
/* Optional: #include "beer.h" or #include "caffeine.h" */

/* 
  * SESSION COMMANDS
  ********* */


/*
  * EXTERNAL FUNCTIONS, RETURN CHAR POINTERS 
  ********* */
char *
doGet(snmp_session *sess, char *oids)
{
	char *buff;
	Packet *req, *rcv;

	req = mkpkt(sess, GetRequest, oids, nil, 0, 0);
	rcv = mkpkt(nil, 0, nil, nil, 0, 0);

	buff = dumppkt(getit(sess, req, rcv));

	freepkt(req);
	freepkt(rcv);
	return buff;
}


char *
doGetNext(snmp_session *sess, char *oids)
{
	char *buff;
	Packet *req, *rcv;

	req = mkpkt(sess, GetNextRequest, oids, nil, 0, 0);
	rcv = mkpkt(nil, 0, nil, nil, 0, 0);

	buff = dumppkt(getit(sess, req, rcv));

	freepkt(req);
	freepkt(rcv);
	return buff;
}


char *
doSet(snmp_session *sess, char *oids, char *values)
{
	char *buff;
	Packet *req, *rcv;

	req = mkpkt(sess, SetRequest, oids, values, 0, 0);
	rcv = mkpkt(nil, 0, nil, nil, 0, 0);

	buff = dumppkt(getit(sess, req, rcv));

	freepkt(req);
	freepkt(rcv);
	return buff;
}


char *
doWalk(snmp_session *sess, char *oids)
{
	char *buff;
	Packet *req, *rcv;

	req = mkpkt(sess, GetNextRequest, oids, nil, 0, 0);
	rcv = mkpkt(nil, 0, nil, nil, 0, 0);

	buff = dumppkt(walkit(sess, req, rcv));

	freepkt(req);
	freepkt(rcv);
	return buff;
}

char *
doBulk(snmp_session *sess, char *oids, int nonrep, int maxrep)
{
	char *buff;
	Packet *req, *rcv;

	req = mkpkt(sess, GetBulkRequest, oids, nil, nonrep, maxrep);
	rcv = mkpkt(nil, 0, nil, nil, 0, 0);

	buff = dumppkt(bulkit(sess, req, rcv));

	freepkt(req);
	freepkt(rcv);
	return buff;
}


/*
  * INTERNAL FUNCTIONS, RETURN PACKETS
  ********* */
Packet *
getit(snmp_session *s, Packet *snd, Packet *rcv)
{
	error(DEBUG, PLUS, "getit()", "Doing a SNMP Get.");

	if ( talk(s, snd, rcv) < 0 ) {
		error(DEBUG, EQUAL, "getit()", "error talking snmp.");
		rcv = nil;
	}

	error(DEBUG, MINUS, "getit()", "SNMP Get Done.");

	return(rcv);
}


Packet *
walkit(snmp_session *sess, Packet *pktreq, Packet *pktrcv)
{
	Packet *tosnd, *torcv, *header, *prelast;
	int errflg = 0;
	char *oidchar;
	int round = 0;
	int goon = 0;

	header = nil;
	prelast = nil;

	error(DEBUG, PLUS, "walkit()", "Doing a SNMP Walk.");

	/* Preparing the first packet to send */
	tosnd = pktreq;
	torcv = pktrcv;

	do {
		if ( talk(sess, tosnd, torcv) < 0 ) {
			error(DEBUG, EQUAL, "walkit()", "Error talking snmp.");
			errflg = 1;
		} else {
			/* Save the packet in a list 
			  * except if is the first response
			  * beacuse is the same packet
			  */
			if (round == 0) {
				/* Preparing the header of the packet list */
				header = pktrcv;
				prelast = pktrcv;
				round = 1;
				goon = 1;
			} else {
				goon = canwalk(prelast, torcv);
				/* If are different */
				if (goon) {
					/* Link the packet */
					linkpkt(header, torcv);
					/* Prepare the next comparation */
					prelast = torcv;
				}
			}

			/* kix Free some memory here */

			/* Get the next oid to ask */
			oidchar = asn2char(torcv->pdu->tuple->oid);
			/* Create the new packet to send*/
			tosnd = mkpkt(sess, GetNextRequest, oidchar, nil, 0, 0);
			/* Create the new packet to recive */
			torcv = mkpkt(nil, 0, nil, nil, 0, 0);
		}
	} while( goon && !errflg );

	error(DEBUG, MINUS, "walkit()", "SNMP Walk done.");
	return header;
}

int
canwalk(Packet *header, Packet *torcv)
{
	char *pkt1c, *pkt2c;

	pkt1c = asn2char(header->pdu->tuple->oid);
	pkt2c = asn2char(torcv->pdu->tuple->oid);

	if (strcmp(pkt1c, pkt2c)) {
		return(1);
	} else {
		return(0);
	}
}


Packet *
setit(snmp_session *s, Packet *req, Packet *rcv)
{
	return getit(s, req, rcv);
}


Packet *
bulkit(snmp_session *s, Packet *req, Packet *rcv)
{	
	error(DEBUG, MINUS, "bulkit()", "Doing a SNMP GetBulk.");

	if ( talk(s, req, rcv) < 0 ) {
		error(DEBUG, EQUAL, "bulkit()", "Error talking snmp.");
		rcv = nil;
	}

	error(DEBUG, MINUS, "bulkit()", "GetBulk Done.");

	return(rcv);
}

/*
  * OTHER FUNCTIONS
  *******/
char * 
dumppkt(Packet *pkt)
{
	char *line, *valchar, *oidchar, *buff, *ptr;
	int n, linelen, len = 0;
	Tuple *tmptpl;
	Packet *tmppkt;

	error(DEBUG, PLUS, "dumppkt()", "Dumping the packets.");

	if (pkt == nil) {
		error(DEBUG, MINUS, "dumppkt()", "The packet is nil.");
		return(nil);
	}

	buff = emalloc9p(0);
	tmppkt = pkt;

	while (tmppkt != nil) {
		error(DEBUG, EQUAL, "dumppkt()", "Dumping a packet.");
		tmptpl = tmppkt->pdu->tuple;

		while (tmptpl != nil) {
			error(DEBUG, EQUAL, "dumppkt()", "Dumping a tuple.");
			oidchar = asn2char(tmptpl->oid);
			valchar = asn2char(tmptpl->value);
			error(DEBUG, EQUAL, "dumppkt()", "OID = %s", oidchar);

			linelen = strlen(oidchar) + strlen(valchar) + 9;
			buff = erealloc9p(buff, len + linelen);
			line = emalloc9p(linelen);
			n = snprint(line, linelen + 9, "[(%s) (%s)]\n", oidchar, valchar);
			ptr = buff + len;
			memcpy(ptr, line, n);
			len += n;

			free(oidchar);
			free(valchar);
			free(line);

			tmptpl = tmptpl->nexttuple;
		}
		tmppkt = tmppkt->nextpkt;
	}

	buff[len] = '\0';

	error(DEBUG, MINUS, "dumppkt()", "Packets dumped.");
	
	return buff;
}

/* TO OBSOLETE */
char *
dosnmp(snmp_session *sess, char *oids, char *values, int type, int n, int m)
{
	char *buff;
	Packet *req, *rcv;

	error(DEBUG, PLUS, "dosnmp()", "Doing an SNMP query.");

	buff = nil;
	req = nil;
	rcv = nil;

/*	req = mkpkt(sess, type, oids, values, n, m);
	rcv = mkpkt(nil, 0, nil, nil, 0, 0); */

//	print("AQUI: %d\n", req->pdu->tuple->value->val->u);

//	if (req != nil) {
		switch(type) {
			case GetRequest:
			case GetNextRequest:
			case SetRequest:
				req = mkpkt(sess, type, oids, values, n, m);
				rcv = mkpkt(nil, 0, nil, nil, 0, 0);
				buff = dumppkt(getit(sess, req, rcv));
				break;
			case GetBulkRequest:
				req = mkpkt(sess, GetNextRequest, oids, values, n, m);
				rcv = mkpkt(nil, 0, nil, nil, 0, 0);
				buff = dumppkt(bulkit(sess, req, rcv));
				break;
		}
//	} else {
//		error(DEBUG, EQUAL, "dosnmp()", "The request packet is empty.");
//	}

	freepkt(req);
	freepkt(rcv);
	error(DEBUG, MINUS, "dosnmp()", "SNMP query done.");
	return buff;
}
