#include <u.h>
#include <libc.h>
#include <bio.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include "../lib/asn1.h"
/* *****************
  * Function file for snmp from/to asn1
  * ***************** */

/* *****************
  * Create Elem'ents
  * ***************** */
Elem
mkelem(int type, Elist *el)
{
	Elem e;

	e.tag.class = type;
	e.tag.num = SEQUENCE;
	e.val.tag = VSeq;
	e.val.u.seqval = el;
	return e;
}

/* *****************
  * Convert a string to a Int
  * ***************** */
Ints *
snmp_mkoid(char *str)
{
	char *f[32];
	int i, n;
	Ints *oid;

	n = getfields(str, f, 32, 0, ".");
	if ( n < 0 ) {
		oid = nil;
	} else {
		/* Gabi, I find a memory leak here */
		oid = (Ints*) emalloc9p(sizeof(Ints) * n);
	
		oid->len = n;
		for(i=0; i<n; i++)
			oid->data[i] = atoi(f[i]);
	}
	return oid;
}


/* *****************
  * Convert an asn string Elem to char
  * ***************** */
char *
asn2char(Elem e)
{
	Value v;
	int i, n, len;
	char *str = nil, *aux, *ptr;
	
	/* chek if i have a valid type before trying to decode it */
	if ( e.tag.class > 30 ) {
		str = smprint("asn2char(): Unknown class type %x and val tag %x", e.tag.class, e.val.tag);
	} else {
		v = e.val;
		
		/* TODO: complete all known types */
		if ( e.tag.class == Universal ) {
			switch(v.tag) {
				case VBigInt:
					str = smprint("BINGINT");
					break;
				case VBitString:
					str = smprint("BITSTRING");
					break;
				case VBool:
					str = smprint("BOOLEAN");
					break;
				case VEOC:
					str = smprint("EOC");
					break;
				case VInt:
					str = smprint("%d", v.u.intval);
					break;
				case VNull:
					str = smprint("NULL");
					break;
				case VReal:
					str = smprint("REAL");
					break;
				case VObjId:
					len = 0;
					str = (char*)emalloc9p(1);
					for(i = 0; i<v.u.objidval->len; i++) {
						aux = smprint("%d%s", v.u.objidval->data[i],(i != (v.u.objidval->len-1)) ? ".":"");
						n = strlen(aux);
						len += n;
						str = (char*)erealloc9p(str,len);
						ptr = str+(len-n);
						memcpy(ptr,aux,n);
						free(aux);
					}
					str = (char*)erealloc9p(str,len+1);
					str[len] = '\0';
					break; 
				case VOctets:
					str = (char*) emalloc9p(v.u.octetsval->len);
					memcpy(str, v.u.octetsval->data, v.u.octetsval->len);
					str = (char*)erealloc9p(str,v.u.octetsval->len+1);
					str[v.u.octetsval->len] = '\0';
					break;
				case VSeq:
					str = smprint("SEQUENCE");
					break;
				case VSet:
					str = smprint("SET");
					break;
				case VString:
					str = smprint("%s", v.u.stringval); break;
					break;
				default:
					str = smprint(str,"Class %d - Type %d is not decoded", e.tag.class, v.tag);
					break;
			}		
		} else {
			str = smprint("asn2char(): Class is not UNIVERSAL %x", e.tag.class);
		}
	}
	return str;
}

/* ----
  * getBulkRequest PDU ASN1
  *	Request-ID INTEGER
  *	Non-Repeaters INTEGER
  *	Max-Repetition INTEGER
  *	Var-Bind-List =SEQUENCE 		
  *
  * OR
  *
  * getNextRequest PDU ASN1				
  *		Request-ID INTEGER			
  *		Error-Status INTEGER			
  *		Error-Index INTEGER			
  *		Var-Bind-List =SEQUENCE 		
  * -----*/
Elist *
mkrequest(uchar type, int reqid, int f1, int f2, Elem el)
{
	Elist *list1, *list2;
	Elem ereq, e1, e2, e3;

	ereq = mkint(reqid);
	e1 = mkint(f1);				/* Element */
	e2 = mkint(f2);				/* Element */
	list1 = mkel(el, nil);			/* New list */
	list1 = mkel(e2, list1);		/* Add */
	list1 = mkel(e1, list1);		/* Add */
	list1 = mkel(ereq, list1);		/* Add */
	e3 = mkelem((int) type, list1);
	list2 = mkel(e3, nil);

	return list2;				// leak
}


/* 
  * the asn.1 thing of libsec stores unknown objects as OCTECT_STRING
  * see ber_decode() function in asn1.h, we will take those raw bytes and
  * will construct a sequence adding the sequence ID and lenght of the
  * bytes encoded  at the begining of the bytes in eepdu, then
  * we decode as a normal sequence
  */

Elem *
pdu2asn(Elem *epdu)
{
	Elem *e;
	int err, ilen;
	uchar *data, *psave, *pdata = nil, seqid=0x30, *seqlen = nil;
	long datalen, len, seqlensize;

	/* encode the size of the sequence */
	datalen = epdu->val.u.octetsval->len;

	if ( datalen < 0x80 ) {
		seqlensize =1;
		seqlen=(uchar*) malloc(seqlensize);
		*seqlen = datalen;
	} else {
		/* First, calculate the len of the secuence */
		psave = pdata;
		int_enc(&pdata, datalen, 1, 1);
		ilen = pdata - psave;
		pdata = psave;
		pdata+=1+ilen;

		/* Create the memory space */
		seqlensize = pdata-seqlen;
		seqlen = (uchar*)malloc(seqlensize);
		pdata = seqlen;

		/* Now, set the data */
		psave = pdata;
		int_enc(&pdata, datalen, 1, 1);
		ilen = pdata - psave;
		pdata = psave;
		*pdata++ = (0x80 | ilen);
		int_enc(&pdata, datalen, 1, 0);
	}

	// make room for the entire object and glue all 
	len = datalen + seqlensize + 1;
	data = (uchar*) emalloc9p(len);
	memset(data, 0, len);
	
	data[0] = seqid; 
	memcpy(&data[1], seqlen, seqlensize);
	memcpy(&data[seqlensize+1], epdu->val.u.octetsval->data, datalen);

	e = (Elem*) malloc(sizeof(Elem));
	err = decode(data, len, e);

	/* Frees */
	free(seqlen);
	free(data);

	if ( err != ASN_OK )
		e = nil;

	return e;
}

/* this is used to compose set-request message */
/* only int and octectstring */
Elem
char2asn(char *data)
{
	/* for normal requests */
	if ( data == nil )
		return Null();

	/* for set-request only octect-strings and integers */
	if ( data[0] < '0' || data[0] > '9' ) {
		return mkoctet((uchar*)data, strlen(data));
	} else {
		return mkint(atoi(data));
	}
}
